﻿using System;
using System.Collections.Generic;
using System.Text;
using gudusoft.gsqlparser;
using gudusoft.gsqlparser.Units;
using System.IO;
using System.Collections;

namespace ScanTable
{
    #region tableInfo
    internal class tableInfo
    {
        public TSourceToken database, schema, table;

        public string getDatabaseName()
        {
            if (database == null)
            {
                return "";
            }
            else
            {
                return database.AsText;
            }
        }

        public string getSchemaName()
        {
            if (schema == null)
            {
                return "";
            }
            else
            {
                return schema.AsText;
            }
        }

        public string getTableName()
        {
            if (table == null)
            {
                return "";
            }
            else
            {
                return table.AsText;
            }
        }

        public string getTableFullName()
        {
            StringBuilder builder = new StringBuilder();
            if (getDatabaseName().Length > 0)
            {
                builder.Append(getDatabaseName()).Append(".");
            }
            if (getSchemaName().Length > 0)
            {
                builder.Append(getSchemaName()).Append(".");
            }
            if (getTableName().Length > 0)
            {
                builder.Append(getTableName());
            }
            return builder.ToString();
        }
    }
    #endregion

    #region columnLocation
    internal enum columnLocation
    {
        ResultColumn, Where, OrderBy, GroupBy, Having, From, Join, Unknown
    }
    #endregion

    public class scantable
    {
        private StringBuilder builder = new StringBuilder();
        private Hashtable tableMap = new Hashtable();

        public String Result
        {
            get { return builder.ToString(); }
        }

        public scantable(string sql, TDbVendor dbVendor)
        {
            TGSqlParser sqlparser = new TGSqlParser(dbVendor);
            sqlparser.SqlText.Text = sql;
            analyzeSQL(sqlparser);
        }

        public scantable(FileInfo file, TDbVendor dbVendor)
        {
            TGSqlParser sqlparser = new TGSqlParser(dbVendor);
            sqlparser.Sqlfilename = file.ToString();
            analyzeSQL(sqlparser);
        }

        private void analyzeSQL(TGSqlParser sqlparser)
        {
            int ret = sqlparser.Parse();

            if (ret != 0)
            {
                builder.AppendLine(sqlparser.ErrorMessages);
                return;
            }
            else
            {
                for (int i = 0; i < sqlparser.SqlStatements.Count(); i++)
                {
                    iterateStmt(sqlparser.SqlStatements[i]);
                }
            }

            IEnumerator iter = tableMap.Keys.GetEnumerator();
            while (iter.MoveNext())
            {
                String key = iter.Current as String;
                builder.Append("Table Name:\t").AppendLine(key);
                Hashtable columnInfos = (Hashtable)tableMap[key];
                if (columnInfos != null && columnInfos.Count > 0)
                {
                    printColumnInfo(builder, columnInfos, columnLocation.ResultColumn, "SELECT LIST:\t");
                    printColumnInfo(builder, columnInfos, columnLocation.From, "FROM:\t\t");
                    printColumnInfo(builder, columnInfos, columnLocation.Where, "WHERE:\t\t");
                    printColumnInfo(builder, columnInfos, columnLocation.GroupBy, "GROUP BY:\t");
                    printColumnInfo(builder, columnInfos, columnLocation.Having, "HAVING:\t\t");
                    printColumnInfo(builder, columnInfos, columnLocation.OrderBy, "ORDER BY:\t");
                    printColumnInfo(builder, columnInfos, columnLocation.Join, "JOIN:\t\t");
                    printColumnInfo(builder, columnInfos, columnLocation.Unknown, "OTHER:\t\t");
                }
                builder.AppendLine();
            }
        }

        protected void iterateStmt(TCustomSqlStatement stmt)
        {
            switch (stmt.SqlStatementType)
            {
                case TSqlStatementType.sstSelect:
                    analyzeSelect((TSelectSqlStatement)stmt);
                    break;
                default:
                    break;
            }
            for (int i = 0; i < stmt.ChildNodes.Count(); i++)
            {
                if (stmt.ChildNodes[i] is TCustomSqlStatement)
                    iterateStmt((TCustomSqlStatement)stmt.ChildNodes[i]);
            }
        }

        private void printColumnInfo(StringBuilder builder, Hashtable columnInfos, columnLocation columnLocation, string hint)
        {
            if (columnInfos.ContainsKey(columnLocation))
            {
                List<String> columnList = (List<String>)columnInfos[columnLocation];
                builder.Append(hint);
                for (int i = 0; i < columnList.Count; i++)
                {
                    builder.Append(columnList[i]);
                    if (i < columnList.Count - 1)
                        builder.Append(", ");
                    else builder.AppendLine();
                }
            }
        }

        private void analyzeSelect(TSelectSqlStatement stmt)
        {

            TSourceTokenList tokenList = stmt.TableTokens;

            TSourceToken st = null;

            for (int i = 0; i < tokenList.Count(); i++)
            {
                tableInfo tableInfo = new tableInfo();
                st = tokenList[i];
                tableInfo.table = st;

                if (st.ParentToken != null)
                {
                    //schema
                    tableInfo.schema = st.ParentToken;

                    if (st.ParentToken.ParentToken != null)
                    {
                        //database
                        tableInfo.database = st.ParentToken.ParentToken;
                    }
                }
                if (!tableMap.ContainsKey(tableInfo.getTableFullName()))
                {
                    tableMap.Add(tableInfo.getTableFullName(), new Hashtable());
                }
                analyzeTableToken(stmt, st, tableInfo);
            }
        }

        private void analyzeTableToken(TSelectSqlStatement stmt, TSourceToken st, tableInfo tableInfo)
        {
            Hashtable columnInfos = (Hashtable)tableMap[tableInfo.getTableFullName()];
            if (st.RelatedToken != null)
            {
                // declared table alias token
                TSourceToken rt = st.RelatedToken;
                TSourceToken rrt = null;
                for (int i = 0; i < rt.RelatedTokens.Count(); i++)
                {
                    rrt = rt.RelatedTokens[i];
                    if (rrt.ChildToken != null)
                    {
                        addColumn(columnInfos, rrt.ChildToken);
                    }
                }
            }

            TSourceToken rtt = null;
            for (int i = 0; i < st.RelatedTokens.Count(); i++)
            {
                // reference table token
                rtt = st.RelatedTokens[i];
                if (rtt.DBObjType == TDBObjType.ttObjField)
                {
                    // get all field tokens link with table token (those token not linked by syntax like tablename.fieldname)
                    // but like this : select f from t
                    addColumn(columnInfos, rtt);
                }
                if (rtt.ChildToken != null)
                {
                    addColumn(columnInfos, rtt.ChildToken);
                }
            }
        }

        private void addColumn(Hashtable columnInfos, TSourceToken st)
        {
            List<String> list;
            if ((st.Location == TLzOwnerLocation.elfieldbyattr)
                || (st.Location == TLzOwnerLocation.elfieldbyexpr)
                )
            {
                list = getColumns(columnInfos, columnLocation.ResultColumn);
            }
            else if (st.Location == TLzOwnerLocation.elWhere
                || st.Location == TLzOwnerLocation.elWhere_not_top_level)
            {
                list = getColumns(columnInfos, columnLocation.Where);
            }
            else if (st.Location == TLzOwnerLocation.elHaving)
            {
                list = getColumns(columnInfos, columnLocation.Having);
            }
            else if (st.Location == TLzOwnerLocation.elgroupby)
            {
                list = getColumns(columnInfos, columnLocation.GroupBy);
            }
            else if (st.Location == TLzOwnerLocation.elorderby)
            {
                list = getColumns(columnInfos, columnLocation.OrderBy);
            }
            else if (st.Location == TLzOwnerLocation.eltableFuncArg)
            {
                list = getColumns(columnInfos, columnLocation.From);
            }
            else if (st.Location == TLzOwnerLocation.eljoinitemcondition)
            {
                list = getColumns(columnInfos, columnLocation.Join);
            }
            else
            {
                list = getColumns(columnInfos, columnLocation.Unknown);
            }
            if (!list.Contains(st.AsText))
                list.Add(st.AsText);
        }

        private List<String> getColumns(Hashtable columnInfos, columnLocation columnLocation)
        {
            if (!columnInfos.ContainsKey(columnLocation))
            {
                columnInfos.Add(columnLocation, new List<String>());
            }
            return (List<String>)columnInfos[columnLocation];
        }

        static void Main(string[] args)
        {
            if (args.Length == 0)
            {
                Console.WriteLine("Usage: {0} <sql script file path> [/o <output file path>] [/t <database type>]", Environment.GetCommandLineArgs()[0]);
                Console.WriteLine("sql script file path: The sql file will be analyzed.");
                Console.WriteLine("/o: Option, write the output stream to the specified file.");
                Console.WriteLine("/t: Option, set the database type. Support oracle,mssql,mysql and db2, the default type is mssql");
                //Console.Read();
                return;
            }

            Array array = Array.CreateInstance(typeof(string), args.Length);

            for (int i = 0; i < array.Length; i++)
            {
                array.SetValue(args[i], i);
            }

            string outputFile = null;

            int index = Array.IndexOf(array, "/o");

            if (index != -1 && args.Length > index + 1)
            {
                outputFile = args[index + 1];
            }

            StreamWriter writer = null;
            if (outputFile != null)
            {
                try
                {
                    writer = new StreamWriter(outputFile);
                    Console.SetOut(writer);
                }
                finally { }
            }

            TDbVendor vendor = TDbVendor.DbVMssql;

            index = Array.IndexOf(array, "/t");

            if (index != -1 && args.Length > index + 1)
            {
                if (args[index + 1].Equals("oracle"))
                {
                    vendor = TDbVendor.DbVOracle;
                }
                else if (args[index + 1].Equals("db2"))
                {
                    vendor = TDbVendor.DbVDB2;
                }
                else if (args[index + 1].Equals("mysql"))
                {
                    vendor = TDbVendor.DbVMysql;
                }
            }

            scantable analysis = new scantable(new FileInfo(args[0]), vendor);
            Console.Write(analysis.Result);
            //if (args.Length <= 1)
            //{
            //    Console.Read();
            //}
            //else
            {
                if (writer != null)
                {
                    writer.Close();
                }
            }

        } //main
    }
}
